#
# The Quick and Dirty Search Engine and Indexer
#
# (C) Copyright 2001 The Internet (Aust) Pty Ltd
# ACN: 082 081 472  ABN: 83 082 081 472
# All Rights Reserved
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Author: Andrew Milton <akm@theinternet.com.au>

import App, Globals, OFS
import string
import time
import md5, base64
import re

from Persistence import Persistent
from Globals import ImageFile, HTMLFile, HTML, MessageDialog, package_home
from Acquisition import Implicit
from OFS.Folder import Folder
from App.Extensions import getObject
from webdav.NullResource import NullResource
from ZPublisher.HTTPResponse import HTTPResponse


# Article:
# id
# wordlist
# wordFreq

def manage_addTextIndexer(self, id, summarySize, SearchMode, ignorePerc, iFreq,title='', REQUEST={}):
	""" Add a new Text Indexer """
	
	obj=TextIndexer(id, summarySize, long(iFreq),title, SearchMode, ignorePerc)
	self._setObject(id, obj)

	if REQUEST:
		return self.manage_main(self, REQUEST)
	return ''

manage_addTextIndexerForm=HTMLFile('manage_addTextIndexerForm', globals())

class TextIndexer(Folder):
	meta_type='Text Indexer'

	manage_indexingOptionsForm=HTMLFile('manage_indexingOptionsForm', globals())
	searchForm=HTMLFile('searchForm', globals())
	searchResults=HTMLFile('searchResults', globals())
	statsPage=HTMLFile('statsPage', globals())
	manage_main=manage_indexingOptionsForm
	searchBox=HTMLFile('searchBox', globals())
	manage_options=(
		{'label':'Indexer', 'action':'manage_indexingOptionsForm'},
		{'label':'Search', 'action':'searchForm'},
		{'label':'Stats', 'action':'statsPage'},
		)
	
	def __init__(self, id, summarySize, iFreq, title='', SearchMode='and', ignorePerc=100.0):
		self.id=id
		self.title=title
		self.iFreq=iFreq
		self.summarySize=summarySize
		self.words={}
		self.ignorePerc=ignorePerc
		self.totalWordFreq=0
		self.letterInWords={}
		self.articles={}
		self.wordInArticles={}
		self.letterFreqs={}
		self.initialiseLetterFreqs()
		self.initialiseLetterInWords()

		crapString=(string.punctuation+
					string.digits+
					string._idmap[0:32]+
					string._idmap[127:])
		
		idMap1=string._idmap
		idMap2=''
		for x in idMap1:
			if x in crapString:
				idMap2=idMap2+' '
			else:
				idMap2=idMap2+x
		self.transMap=idMap2
		self.SearchMode=SearchMode

	def initialiseLetterFreqs(self):
		for c in string.lowercase:
			self.letterFreqs[c]=0
		self._p_changed=1
		
	def initialiseLetterInWords(self):
		for c in string.lowercase:
			self.letterInWords[c]=[]
		self._p_changed=1			
			
	def addWordToLetterIndex(self, word, windex):
		letters={}

		# String to set in one easy move..
		for c in word:
			letters[c]=1
			try:
				self.letterFreqs[c]+=1
			except:
				print "Fail %d"%(ord(c))

		for k in letters.keys():
			self.letterInWords[k].append(windex)
		self._p_changed=1			

	def getDigest(self, text):
		digest=md5.new()
		digest.update(text)
		digest=digest.digest()
		digest=string.strip(base64.encodestring(digest))
		return digest

	def stripHtmlTags(self, data):
		# Remove comments first, or '>' in
		# comments will be interpreted as
		# end of (comment) tag.
		text = re.sub('<!--.*?-->', '', data) 

		text = re.sub('<.*?>', '', text)
		return text
		
	def addArticleToIndexer(self, articleId, articleBody, artType='url'):
		""" Index something """
			
		digest=None
		now=time.time()
		if self.articles.has_key(articleId):
			if self.articles[articleId]['nextIndex'] < now:
				return
			digest=self.getDigest(articleBody)
			# Content has changed?
			if self.articles[articleId]['digest'] == digest:
				# No
				return

		if not digest:
			digest=self.getDigest(articleBody)
		
		artWords=[]
		article={}
		spaces=''

		# Strip HTML from URL based content

		if artType=='url':
			articleBody=self.stripHtmlTags(articleBody)
		
		# Strip all punctuation, digits, and crap that's not printable
		# -- we want words with letters in them only at this point

		sBody=string.translate(articleBody, self.transMap)
		
		wordList=string.split(sBody)
		del sBody # don't need a full copy of the article anymore..
		for word in wordList:
			self.totalWordFreq+=1			
			word=string.lower(word)
			if self.words.has_key(word):
				windex=self.words[word]['index']
			else:
				windex=len(self.words)
				self.words[word]={'index':windex,
								  'freq':0}
				self.addWordToLetterIndex(word, windex)
			self.words[word]['freq']+=1
			if windex not in artWords:
				artWords.append(windex)


		self.wordInArticles[articleId]=artWords
		self.articles[articleId]={
			'digest':digest,
			'nextIndex':time.time()+self.iFreq,
			'type':artType}

		# Keep a summary
		if self.summarySize:
			self.articles[articleId]['summary']=articleBody[:self.summarySize]
		else:
			# Simpler to leave something rather than have to check for a key
			# every time
			self.articles[articleId]['summary']='' 
		self._p_changed=1

	def manage_indexURL(self, url, REQUEST={}):
		""" Index a URL """
		import urllib
		(fname, headers)=urllib.urlretrieve(url)
		fp=open(fname)
		data=fp.readlines()
		data=string.join(data,'\n')
		fp.close()
		self.addArticleToIndexer(url, data,'url')

		if REQUEST:
			return self.manage_main(self, REQUEST)
		return ''

	def manage_indexObject(self, id, file='', REQUEST={}):
		""" Index a File """
		seek=file.seek
		read=file.read
		seek(0,2)
		size=file.tell()
		seek(0)
		data=read(size)
		self.addArticleToIndexer(id, data, 'file')

		if REQUEST:
			return self.manage_main(self, REQUEST)
		return ''


	def findWordsInArticles_and(self, wordList):
		articles=[]
		tmpArticles=[]
		for word in wordList:
			word=string.lower(word)
			if self.words.has_key(word):
				if self.ignorePerc < 100.0:
					if ((float(self.words[word]['freq'])/
						float(self.totalWordFreq))*100) > self.ignorePerc:
						print "%s exceeds threshold of %f%% (%f%%)"%(
							word, self.ignorePerc,
							((float(self.words[word]['freq'])/
							  float(self.totalWordFreq))*100))
						continue
				windex=self.words[word]['index']
			else:
				return


			if not tmpArticles:
				tmpArticles=self.wordInArticles.keys()
			else:
				tmpAricles=articles
			articles=[]
			
			for a in tmpArticles:
				if windex in self.wordInArticles[a]:
					articles.append(a)

		# We could change this to add relevence by
		# boosting things up and down the order based
		# on the total freqs of the words in the article etc..
		# -- later
		return articles
		
	# This can be pretty damn quickly optimised for multiple words
	# By starting with the word with the lowest frequency
	# If we ever get a word with no matches we leave early anyway (in and mode)

	def findWordsInArticles(self, wordList, SearchMode=''):
		""" Return a List of Articles containing word list """

		# Did the override the default search Mode?
		if not SearchMode:
			SearchMode=self.SearchMode

		if SearchMode=='and':
			# Use optimised 'and' search
			return self.findWordsInArticles_and(wordList)
		
		articles=[]
		for word in wordList:
			word=string.lower(word)
			if self.words.has_key(word):
				if self.ignorePerc < 100.0:
					if ((float(self.words[word]['freq'])/
						float(self.totalWordFreq))*100) > self.ignorePerc:
						print "%s exceeds threshold of %f%% (%f%%)"%(
							word, self.ignorePerc,
							((float(self.words[word]['freq'])/
							  float(self.totalWordFreq))*100))
						continue
				windex=self.words[word]['index']
			else:
					continue

			for a in self.wordInArticles.keys():
				if windex in self.wordInArticles[a]:
					if a not in articles:
						articles.append(a)

		# We could change this to add relevence by
		# boosting things up and down the order based
		# on the total freqs of the words in the article etc..
		# -- later
		return articles

	def printStatistics(self):
		print "Number of articles: %d"%(len(self.wordInArticles))
		print "Number of unique words: %d"%(len(self.words))
		print "Word Frequencies"
		for w in self.words.keys():
			print "%s: %d"%(w, self.words[w]['freq'])

